#include "StdAfx.h"
#include "d3d_state_mgr.h"

#include <assert.h>

namespace d3d8_state_mgr {

enum
{
	RENDER_STATE_MAX_NUM = 256,
};

static struct StateVersion
{
	dword version;
	dword ary_renderState[RENDER_STATE_MAX_NUM];	
	dword stream;
	dword indices;
	dword texture;
	dword shader;

	StateVersion()
	{memset(this, 0, sizeof(*this));}
} gs_stateVersion;

static struct StateCache
{	
	IDirect3DVertexBuffer8* pkStream;
	IDirect3DIndexBuffer8*	pkIndices;
	IDirect3DTexture8*		pkTexture;	

	dword ary_renderState[RENDER_STATE_MAX_NUM];	
	dword streamStride;
	dword vertexStartIndex;
	dword shader;	

	D3DMATRIX matProj;

	StateCache()
	{memset(this, 0, sizeof(*this));}
} gs_stateCache;

struct StateItem_SetRenderState {
	D3DRENDERSTATETYPE e_type;
	dword value;
};

struct StateItem_SetTexture
{
	uint stageIndex;
	IDirect3DTexture8* pkTexture;
};

struct StateItem_SetStreamSource
{
	uint streamIndex;
	IDirect3DVertexBuffer8* pkStream;
	uint vertexStride;
};

struct StateItem_SetIndices
{
	IDirect3DIndexBuffer8* pkIndices;
	uint vertexStartIndex;
};

struct StateItem_SetVertexShader
{
	dword handle;
};

struct StateItem
{
	void (*apply)(const StateItem& c_rarg);

	union {
		StateItem_SetRenderState		renderState;
		StateItem_SetStreamSource		stream;
		StateItem_SetIndices			indices;
		StateItem_SetVertexShader		shader;
		StateItem_SetTexture			texture;
	};
	
	StateItem()
	{this->apply = Apply_DoNothing;}

	static void Apply_DoNothing(const StateItem& c_rstateItem)
	{}

	static void Apply_SetStreamSource(const StateItem& c_rstateItem)
	{SetStreamSource(c_rstateItem.stream.streamIndex, c_rstateItem.stream.pkStream, c_rstateItem.stream.vertexStride);}

	static void Apply_SetIndices(const StateItem& c_rstateItem)
	{SetIndices(c_rstateItem.indices.pkIndices, c_rstateItem.indices.vertexStartIndex);}

	static void Apply_SetTexture(const StateItem& c_rstateItem)
	{SetTexture(c_rstateItem.texture.stageIndex, c_rstateItem.texture.pkTexture);}

	static void Apply_SetVertexShader(const StateItem& c_rstateItem)
	{SetVertexShader(c_rstateItem.shader.handle);}

	static void Apply_SetRenderState(const StateItem& c_rstateItem)
	{SetRenderState(c_rstateItem.renderState.e_type, c_rstateItem.renderState.value);}
};
	
static class CStateStack 
{
public:
	CStateStack() : m_pos(0)
	{}

	int GetSize()
	{return m_pos;}

	void Push(const StateItem& c_rarg)
	{
		assert(m_pos<STATESTACK_SIZE);
		m_astateItem[m_pos++] = c_rarg;
	}
	void Flush(int pos)
	{
		if (pos < 0)
			return;
		
		while (m_pos > pos)
		{
			const StateItem& rstateItem = m_astateItem[--m_pos];
			rstateItem.apply(rstateItem);		
		}	
	}
	
private:
	enum
	{
		STATESTACK_SIZE = 128
	};

private:
	StateItem m_astateItem[STATESTACK_SIZE];
	int m_pos;

} gs_kStack;

static IDirect3DDevice8* gs_pkd3dDev = NULL;

static void __BackupVertexShader()
{
	StateItem stateItem;
	stateItem.apply = StateItem::Apply_SetVertexShader;
	stateItem.shader.handle = GetVertexShader();
	gs_kStack.Push(stateItem);	
}

static void __BackupStreamSource(uint streamIndex)
{
	assert(streamIndex == 0);
	StateItem stateItem;
	stateItem.apply = StateItem::Apply_SetStreamSource;
	stateItem.stream.streamIndex = streamIndex;
	stateItem.stream.pkStream = GetStreamSource(streamIndex, &stateItem.stream.vertexStride);	 
	gs_kStack.Push(stateItem);	
}


static void __BackupIndices()
{
	StateItem stateItem;
	stateItem.apply = StateItem::Apply_SetIndices;
	stateItem.indices.pkIndices = GetIndices(&stateItem.indices.vertexStartIndex);
	gs_kStack.Push(stateItem);
}

static void __BackupTexture(uint stageIndex)
{
	assert(stageIndex == 0);
	StateItem stateItem;
	stateItem.apply = StateItem::Apply_SetTexture;
	stateItem.texture.stageIndex = stageIndex;
	stateItem.texture.pkTexture = GetTexture(stageIndex);		
	gs_kStack.Push(stateItem);	
}

static void __BackupRenderState(D3DRENDERSTATETYPE e_type)
{
	if (e_type >= RENDER_STATE_MAX_NUM)
	{
		assert(!"SetRenderState e_type is OVERFLOW");
		return;
	}		

	StateItem stateItem;
	stateItem.apply = StateItem::Apply_SetRenderState;
	stateItem.renderState.e_type=e_type;
	stateItem.renderState.value=GetRenderState(e_type);	
	gs_kStack.Push(stateItem);	
}

void SetDefaultState()
{
	IDirect3DDevice8& rkd3dDev = GetDeviceRef();	
	
	rkd3dDev.SetVertexShader(D3DFVF_XYZ);	
	rkd3dDev.SetStreamSource(0, NULL, 0);
	rkd3dDev.SetTexture(0, NULL);	
	rkd3dDev.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);	
	rkd3dDev.SetRenderState(D3DRS_LIGHTING, FALSE);
	rkd3dDev.SetRenderState(D3DRS_ZENABLE, TRUE);
	rkd3dDev.SetIndices(NULL, 0);

	SetVertexShader(D3DFVF_XYZ);
	SetStreamSource(0, NULL, 0);
	SetTexture(0, NULL);
	SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);		
	SetRenderState(D3DRS_LIGHTING, FALSE);		
	SetRenderState(D3DRS_ZENABLE, TRUE);
	SetIndices(NULL, 0);
}

void SetTransform(D3DTRANSFORMSTATETYPE e_transType, const D3DMATRIX* c_pmatTrans)
{
	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetTransform(e_transType, c_pmatTrans);	
}

void DrawPrimitive(D3DPRIMITIVETYPE e_priType, uint vertexStartIndex, uint priCount)
{
	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.DrawPrimitive(e_priType, vertexStartIndex, priCount);
}

void DrawIndexedPrimitive(D3DPRIMITIVETYPE e_priType, uint vertexMinIndex, uint vertexCount, uint vertexStartIndex, uint uPriCount)
{
	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.DrawIndexedPrimitive(e_priType, vertexMinIndex, vertexCount, vertexStartIndex, uPriCount);
}

void SetRenderState(D3DRENDERSTATETYPE e_type, dword value)
{
	if (e_type >= RENDER_STATE_MAX_NUM)
	{
		assert(!"SetRenderState e_type is OVERFLOW");
		return;
	}	
	if (value == gs_stateCache.ary_renderState[e_type])
		return;

	if (gs_stateVersion.version != gs_stateVersion.ary_renderState[e_type])
	{
		gs_stateVersion.ary_renderState[e_type] = gs_stateVersion.version;
		__BackupRenderState(e_type);
	}

	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetRenderState(e_type, value);
	
	gs_stateCache.ary_renderState[e_type]=value;
}

void SetTexture(uint stageIndex, IDirect3DTexture8* pkTexture)
{
	assert(stageIndex == 0);
	if (gs_stateCache.pkTexture == pkTexture)
		return;

	if (gs_stateVersion.version != gs_stateVersion.texture)	
	{
		gs_stateVersion.texture = gs_stateVersion.version;
		__BackupTexture(stageIndex);
	}

	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetTexture(stageIndex, pkTexture);

	gs_stateCache.pkTexture = pkTexture;
}

void SetStreamSource(uint streamIndex, IDirect3DVertexBuffer8* pkStream, uint vertexStride)
{
	assert(streamIndex == 0);
	if (gs_stateCache.pkStream == pkStream)
		return;	

	if (gs_stateVersion.version != gs_stateVersion.stream)	
	{
		gs_stateVersion.stream = gs_stateVersion.version;
		__BackupStreamSource(streamIndex);
	}

	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetStreamSource(streamIndex, pkStream, vertexStride);

	gs_stateCache.pkStream = pkStream;
}

void SetIndices(IDirect3DIndexBuffer8* pkIndices, uint vertexStartIndex)
{
	if (gs_stateCache.pkIndices == pkIndices)
		if (gs_stateCache.vertexStartIndex == vertexStartIndex)
			return;

	if (gs_stateVersion.version != gs_stateVersion.indices)	
	{
		gs_stateVersion.indices = gs_stateVersion.version;
		__BackupIndices();
	}

	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetIndices(pkIndices, vertexStartIndex);

	gs_stateCache.pkIndices = pkIndices;
	gs_stateCache.vertexStartIndex = vertexStartIndex;
}

void SetVertexShader(dword shader)
{
	if (gs_stateCache.shader == shader)
		return;

	if (gs_stateVersion.version != gs_stateVersion.shader)	
	{
		gs_stateVersion.shader = gs_stateVersion.version;
		__BackupVertexShader();
	}

	IDirect3DDevice8& rkd3dDev = GetDeviceRef();
	rkd3dDev.SetVertexShader(shader);

	gs_stateCache.shader = shader;
}

IDirect3DTexture8* GetTexture(uint stageIndex)
{
	assert(stageIndex == 0);
	return gs_stateCache.pkTexture;
}

IDirect3DIndexBuffer8* GetIndices(uint* pvertexStartIndex)
{
	*pvertexStartIndex = gs_stateCache.vertexStartIndex;
	return gs_stateCache.pkIndices;
}


IDirect3DVertexBuffer8* GetStreamSource(uint streamIndex, uint* pvertexStride)
{
	assert(streamIndex == 0);
	*pvertexStride = gs_stateCache.streamStride;
	return gs_stateCache.pkStream;
}


dword GetRenderState(D3DRENDERSTATETYPE e_type)
{
	if (e_type >= RENDER_STATE_MAX_NUM)
	{
		assert(!"GetRenderState e_type is OVERFLOW");
		return 0;
	}	
	return gs_stateCache.ary_renderState[e_type];
}

dword GetVertexShader()
{
	return gs_stateCache.shader;
}

IDirect3DDevice8& GetDeviceRef()
{
	assert(NULL != gs_pkd3dDev);
	return *gs_pkd3dDev;
}

IDirect3DDevice8* GetDevicePtr()
{
	assert(NULL != gs_pkd3dDev);
	return gs_pkd3dDev;
}

int	BackupStateBlock()
{	
	++gs_stateVersion.version;
	return gs_kStack.GetSize();
}

void RestoreStateBlock(int handle)
{	
	gs_kStack.Flush(handle);
}

void Startup(IDirect3DDevice8* pkd3dDev)
{
	assert(NULL != pkd3dDev);
	gs_pkd3dDev = pkd3dDev;
	gs_pkd3dDev->AddRef();

	SetDefaultState();
}

void Cleanup()
{
	if (gs_pkd3dDev)
		gs_pkd3dDev->Release();

	gs_pkd3dDev = NULL;
}

} // d3d8_state_mgr