#include "StdAfx.h"
#include "d3d_state_mgr.h"
#include "d3d_vbring.h"

#pragma comment(lib, "devil.lib")
#include <il/il.h>

#pragma comment(lib, "d3dx8.lib")
#include <d3dx8.h>

#include <map>

#include <eter/string_utils>

namespace test_image {

using namespace eter;
using namespace std;

struct Size
{
	float width;
	float height;
};

struct Vertex
{
	enum
	{
		FVF = D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_SPECULAR|D3DFVF_TEX1
	};
	float x, y, z, rhw;
	DWORD diffuse;
	DWORD specular;
	float u, v;
};

class ITexture : public smart_pointer_object<ITexture>
{
public:
	static ITexture* NewTexture(const char* c_fileName);

public:
	virtual ~ITexture()
	{}

	virtual void Apply();
	
	virtual const Size& GetImageSize()
	{
		static Size s_imageDefaultSize = {100.0f, 100.0f};
		return s_imageDefaultSize;
	}

	virtual const Size& GetScaleSize()
	{
		static Size s_scaleDefaultSize = {1.0f, 1.0f};
		return s_scaleDefaultSize;
	}
};

class CTexture : public ITexture
{
public:
	CTexture(IDirect3DTexture8* pkd3dTex, const Size& c_imageSize, const Size& c_scaleSize) : 
		m_pkd3dTex(pkd3dTex), m_imageSize(c_imageSize), m_scaleSize(c_scaleSize)
	{}

	virtual ~CTexture()
	{
		if (m_pkd3dTex)
			m_pkd3dTex->Release();
	}

	void Apply()
	{
		d3d8_state_mgr::SetTexture(0, m_pkd3dTex);
	}

	const Size& GetImageSize()
	{return m_imageSize;}

	const Size& GetScaleSize()
	{return m_scaleSize;}

private:
	IDirect3DTexture8* m_pkd3dTex;
	Size m_imageSize;
	Size m_scaleSize;
};

static uint __GetNear2Pow(uint orig)
{
	if (orig<=256)
		return 256;
	else if (orig<=512)
		return 512;
	return 1024;
}

ITexture* ITexture::NewTexture(const char* c_fileName)
{
	ILuint handle;
	ilGenImages(1, &handle);
	ilBindImage(handle);
	if (!ilLoad(IL_TYPE_UNKNOWN, (ILstring)c_fileName))
	{
		ilDeleteImages(1, &handle);
		return ITexture::None();
	}

	uint imageWidth = ilGetInteger(IL_IMAGE_WIDTH);
	uint imageHeight = ilGetInteger(IL_IMAGE_HEIGHT);

	uint textureWidth = __GetNear2Pow(imageWidth);
	uint textureHeight = __GetNear2Pow(imageHeight);

	IDirect3DDevice8& rkd3dDev = d3d8_state_mgr::GetDeviceRef();

	IDirect3DTexture8* pkd3dNewTex;
	if (FAILED(
		rkd3dDev.CreateTexture(textureWidth, textureHeight, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &pkd3dNewTex)
	))
	{
		ilDeleteImages(1, &handle);
		return ITexture::None();
	}

	ilConvertImage(IL_BGRA, IL_BYTE);

	D3DLOCKED_RECT d3dLockRect;
	RECT copyRect = {0, 0, imageWidth, imageHeight};
	if (FAILED(
		pkd3dNewTex->LockRect(0, &d3dLockRect, &copyRect, 0)
	))
	{
		ilDeleteImages(1, &handle);
		return ITexture::None();
	}

	dword* pdwSrcLine = (dword*)ilGetData();
	dword* pdwDstLine = (dword*)d3dLockRect.pBits; 
	uint copySize = imageWidth * sizeof(dword);
	uint srcPitch = imageWidth;
	uint dstPitch = d3dLockRect.Pitch / sizeof(dword);
	for (uint i = 0; i != imageHeight; ++i, pdwDstLine += dstPitch, pdwSrcLine += srcPitch)
		memcpy(pdwDstLine, pdwSrcLine, copySize);
	pkd3dNewTex->UnlockRect(0);

	ilDeleteImages(1, &handle);
	
	Size imageSize;
	imageSize.width		= imageWidth;
	imageSize.height	= imageHeight;

	Size scaleSize;
	scaleSize.width		= imageWidth / float(textureWidth);
	scaleSize.height	= imageHeight / float(textureHeight);
	return new CTexture(pkd3dNewTex, imageSize, scaleSize);
}

void ITexture::Apply()
{
	d3d8_state_mgr::SetTexture(0, NULL);
}


typedef smart_pointer<ITexture> SPITexture;

typedef map<dword, SPITexture> Map_SPITexture;

static Map_SPITexture gs_map_spkTexture;
static smart_pointer<IVertexBufferRing> gs_spkVBRing;

ITexture* __LoadTexture(const char* c_fileName)
{
	dword hash = string_hash(c_fileName);
	Map_SPITexture::iterator f = gs_map_spkTexture.find(hash);
	if (gs_map_spkTexture.end() != f)
		return f->second;

	ITexture* vpkNewTexture = ITexture::NewTexture(c_fileName);	
	gs_map_spkTexture.insert(Map_SPITexture::value_type(hash, vpkNewTexture));
	return vpkNewTexture;
}

void RenderImage(float sx, float sy, const char* c_fileName)
{
	ITexture* vpkTexture = __LoadTexture(c_fileName);
	vpkTexture->Apply();

	const Size& c_rimageSize = vpkTexture->GetImageSize();
	sx -= 0.5f;
	sy -= 0.5f;
	const float sz = 0.0f;
	const float ex = sx + c_rimageSize.width;
	const float ey = sy + c_rimageSize.height;

	const Size& c_rscaleSize = vpkTexture->GetScaleSize();
	const float su = 0.0f;
	const float sv = 0.0f;
	const float eu = c_rscaleSize.width;
	const float ev = c_rscaleSize.height;
	
	
	const float		rhw = 1.0f;

	const uint		VERTEX_NUM = 4;
	const Vertex	ary_vertex[VERTEX_NUM] = {
		{sx, sy, sz, rhw, 0xffffffff, 0xffffffff, su, sv},
		{sx, ey, sz, rhw, 0xffffffff, 0xffffffff, su, ev},
		{ex, sy, sz, rhw, 0xffffffff, 0xffffffff, eu, sv},
		{ex, ey, sz, rhw, 0xffffffff, 0xffffffff, eu, ev},
	};

	d3d8_state_mgr::SetVertexShader(Vertex::FVF);

	IVertexBufferRing& rkVBRing = *gs_spkVBRing;
	if (!rkVBRing.SetVertices(sizeof(ary_vertex), ary_vertex, sizeof(Vertex)))
		return;

	d3d8_state_mgr::DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
}

void Startup()
{
	ilInit();
		
	gs_spkVBRing = IVertexBufferRing::NewVertexBufferRing(Vertex::FVF, sizeof(Vertex) * 4, 16);
}

void Cleanup()
{
	gs_map_spkTexture.clear();
	gs_spkVBRing = NULL;
}

} // test_image

