#include "StdAfx.h"

#include <eter/text_cache.h>

#include <vector>
#include <algorithm>

#include "d3d_state_mgr.h"
#include "d3d_vbring.h"

namespace text_cache {

using namespace std;
using namespace eter;

struct Vertex
{
	enum
	{
		FVF = D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_SPECULAR|D3DFVF_TEX1,
	};
	float x, y, z, rhw;
	DWORD diffuse;
	DWORD specular;
	float u, v;
};

static smart_pointer<IVertexBufferRing> gs_spkVBRing;

class CTextCache : public ITextCache
{
public:
	void Print(const char* src, uint srcLen)
	{
		if (srcLen < 256)
		{
			wchar_t dst[256];
			int dstLen = MultiByteToWideChar(m_cursor.codePage, 0, src, strlen(src), dst, sizeof(dst)-1);
			Printw(dst, dstLen);
		}
		else if (srcLen < 1024)
		{
			wchar_t dst[1024];
			int dstLen = MultiByteToWideChar(m_cursor.codePage, 0, src, strlen(src), dst, sizeof(dst)-1);
			Printw(dst, dstLen);
		}
		else
		{
			syserr("CTextCache::Print - TOO LONG TEXT");
		}
	}
	void Printz(const char* text)
	{
		const uint textLen = strlen(text);
		Print(text, textLen);
	}
	void Printw(const wchar_t* text, uint textLen)
	{
		SIZE textSize;
		if (!GetTextExtentPoint32W(m_dib.hDC, text, textLen, &textSize))
		{
			syserr("CTextCache::Printw/GetTextExtentPoint32W");
			return;
		}

		if (__Cursor_IsTextSizeTooLargeWidth(textSize))
		{
			syserr("CTextCache::Printw/__Cursor_IsTextSizeTooLargeWidth");
			return;
		}

		if (__Cursor_IsTextSizeLineEnd(textSize))
		{
			__Cursor_NewLine();
		}
	
		TextOutW(m_dib.hDC, m_cursor.x, m_cursor.y, text, textLen);

		__WorkingBitmap_UpdateTexture(textSize);

		__Cursor_Next(textSize);		
	}
	void RenderWorkingBitmap(HDC hDstDC, int x, int y)
	{
		SetDIBitsToDevice(
			hDstDC, x, y, m_cursor.maxWidth, m_cursor.maxHeight,
			0, 0, 0,
			m_cursor.maxHeight,
			m_dib.base,
			&m_dib.bmi,
			DIB_RGB_COLORS
			);
	}
	void RenderWorkingTexture(float x, float y)
	{
		const float		sx = x;		
		const float		sy = y;
		const float		sz = 0.0f;
		const float		ex = x + __DIB_GetWidth();
		const float		ey = y + __DIB_GetHeight();
		const float		rhw = 1.0f;

		const float		su = 0.0f;
		const float		sv = 0.0f;
		const float		eu = 1.0f;
		const float		ev = 1.0f;			
		
		const uint		RECT_VERTEX_NUM = 4;
		const Vertex	ary_vertex[RECT_VERTEX_NUM] = {
			{sx, sy, sz, rhw, 0xffffffff, 0xffffffff, su, sv},
			{sx, ey, sz, rhw, 0xffffffff, 0xffffffff, su, ev},
			{ex, sy, sz, rhw, 0xffffffff, 0xffffffff, eu, sv},
			{ex, ey, sz, rhw, 0xffffffff, 0xffffffff, eu, ev},
		};

		d3d8_state_mgr::SetTexture(0, m_pkd3dTopTex);
		
		d3d8_state_mgr::SetVertexShader(Vertex::FVF);
		
		IVertexBufferRing& rkVBRing = *gs_spkVBRing;
		if (!rkVBRing.SetVertices(sizeof(ary_vertex), ary_vertex, sizeof(Vertex)))
			return;
		
		d3d8_state_mgr::DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);		
	}


public:
	CTextCache()
	{
		__DIB_Init();
		__TextureList_Init();
		__Cursor_Init();
	}
	virtual ~CTextCache()
	{
		Destroy();

		assert(__TextureList_IsEmpty());
		assert(__DIB_IsEmpty());		
	}
	bool Create(HDC hSrcDC, uint size, uint codePage)
	{
		if (!__DIB_Create(hSrcDC, size, size, 32))
			return false;

		m_cursor.codePage = codePage;
		m_cursor.maxWidth = size;
		m_cursor.maxHeight = size;

		__WorkingBitmap_BuildTexture();
		return true;
	}
	void Destroy()
	{
		__TextureList_Destroy();
		__DIB_Destroy();
	}

private:
	struct Cursor
	{
		uint x;
		uint y;
		uint maxWidth;
		uint maxHeight;
		uint curLineHeight;
		uint codePage;
	} m_cursor;

	void __Cursor_Init()
	{
		memset(&m_cursor, 0, sizeof(m_cursor));
	}
	bool __Cursor_IsTextSizeTooLargeWidth(const SIZE& c_textSize)
	{
		return c_textSize.cx > m_cursor.maxWidth;
	}
	bool __Cursor_IsTextSizeLineEnd(const SIZE& c_textSize)
	{
		return m_cursor.x + c_textSize.cx > m_cursor.maxWidth;
	}
	void __Cursor_Next(const SIZE& c_textSize)
	{
		m_cursor.curLineHeight = max(m_cursor.curLineHeight, c_textSize.cy);
		m_cursor.x += c_textSize.cx;
	}
	void __Cursor_NewLine()
	{
		m_cursor.y += m_cursor.curLineHeight;
		if (m_cursor.y > m_cursor.maxHeight)
		{
			__WorkingBitmap_ChangeTexture();
			m_cursor.y = 0;
		}
		m_cursor.x = 0;
		m_cursor.curLineHeight = 0;
	}

private:	
	struct DIB 
	{
		HDC			hDC;
		HBITMAP		hBmp;
		BITMAPINFO	bmi;

		dword*		base;
	} m_dib;

private:
	bool __DIB_IsEmpty()
	{
		if (m_dib.hDC && m_dib.hBmp && m_dib.base)
			return false;

		return true;
	}
	void __DIB_Init()
	{
		memset(&m_dib, 0, sizeof(m_dib));
	}
	void __DIB_Destroy()
	{
		if (m_dib.hBmp) 
			DeleteObject(m_dib.hBmp);			

		if (m_dib.hDC)	
			DeleteDC(m_dib.hDC);

		__DIB_Init();
	}
	bool __DIB_Create(HDC hSrcDC, uint width, uint height, uint colorBits)
	{
		memset(&m_dib.bmi.bmiHeader, 0, sizeof(BITMAPINFOHEADER));
		m_dib.bmi.bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
		m_dib.bmi.bmiHeader.biWidth       = +width;
		m_dib.bmi.bmiHeader.biHeight      = -height;
		m_dib.bmi.bmiHeader.biPlanes      = 1;			
		m_dib.bmi.bmiHeader.biBitCount    = colorBits;
		m_dib.bmi.bmiHeader.biCompression = BI_RGB;

		m_dib.hDC = CreateCompatibleDC(hSrcDC);
		if (!m_dib.hDC)
			return false;
		
		m_dib.hBmp = CreateDIBSection(m_dib.hDC, &m_dib.bmi, DIB_RGB_COLORS, (void**)&m_dib.base, NULL, 0);
		if (!m_dib.hBmp)
			return false;
		
		SelectObject(m_dib.hDC, m_dib.hBmp);
		return true;
	}
	uint __DIB_GetWidth()
	{
		return abs(m_dib.bmi.bmiHeader.biWidth);
	}
	uint __DIB_GetHeight()
	{
		return abs(m_dib.bmi.bmiHeader.biHeight);
	}

private:	
	void __WorkingBitmap_BuildTexture()
	{
		if (m_pkd3dTopTex)
		{
			syserr("__WorkingBitmap_BuildTexture - TOP TEXTURE ALREADY EXIST");
			return;
		}

		uint bmpWidth	= __DIB_GetWidth();
		uint bmpHeight = __DIB_GetHeight();

		IDirect3DDevice8& rkd3dDev = d3d8_state_mgr::GetDeviceRef();

		if (FAILED(
			rkd3dDev.CreateTexture(bmpWidth, bmpHeight, 1, 0, D3DFMT_A4R4G4B4, D3DPOOL_MANAGED, &m_pkd3dTopTex)
		)) 
		{
			syserr("__WorkingBitmap_BuildTexture/IDirect3DDevice8::CreateTexture - FAILURE");
			return;		
		}

		return;
	}	
	void __WorkingBitmap_UpdateTexture(const SIZE& c_textSize)
	{
		if (!m_pkd3dTopTex)
		{
			syserr("__WorkingBitmap_UpdateTexture - TOP TEXTURE IS NULL");
			return;
		}

		RECT rcText = {m_cursor.x, m_cursor.y, m_cursor.x + c_textSize.cx, m_cursor.y + c_textSize.cy};

		D3DLOCKED_RECT d3dLockedRect;			
		if (FAILED(
			m_pkd3dTopTex->LockRect(0, &d3dLockedRect, &rcText, 0)
		))
		{
			syserr("__WorkingBitmap_UpdateTexture - IDirect3DTexture8::LockRect FAILURE");
			return;
		}
		
		uint dstPitch = d3dLockedRect.Pitch / sizeof(word);
		word* dstLine = (word*)d3dLockedRect.pBits;
		
		uint srcPitch = abs(m_dib.bmi.bmiHeader.biWidth);
		dword* srcLine = m_dib.base + srcPitch * rcText.top + rcText.left;

		uint copyHeight = c_textSize.cy;
		uint copyWidth = c_textSize.cx;
		for (uint y = 0; y != copyHeight; ++y, dstLine += dstPitch, srcLine += srcPitch)
			for (uint x = 0; x != copyWidth; ++x)
				dstLine[x] = srcLine[x];

		m_pkd3dTopTex->UnlockRect(0);
	}
	void __WorkingBitmap_ChangeTexture()
	{
		if (!m_pkd3dTopTex)
		{
			syserr("__WorkingBitmap_ChageTexture - TOP TEXTURE IS NULL");
			return;
		}

		m_vec_pkd3dOldTex.push_back(m_pkd3dTopTex);		
		m_pkd3dTopTex = NULL;

		__WorkingBitmap_BuildTexture();
	}

private:
	void __TextureList_Init()
	{
		m_pkd3dTopTex = NULL;
	}
	bool __TextureList_IsEmpty()
	{
		if (m_pkd3dTopTex)
			return false;

		return m_vec_pkd3dOldTex.empty();
	}
	void __TextureList_Destroy()
	{
		__Texture_Release(m_pkd3dTopTex);
		m_pkd3dTopTex = NULL;

		for_each(m_vec_pkd3dOldTex.begin(), m_vec_pkd3dOldTex.end(), __Texture_Release);
		m_vec_pkd3dOldTex.clear();
	}
	static void __Texture_Release(IDirect3DTexture8* pkd3dTex)
	{
		if (pkd3dTex)
			pkd3dTex->Release();
	}

private:
	IDirect3DTexture8* m_pkd3dTopTex;
	vector<IDirect3DTexture8*> m_vec_pkd3dOldTex;

private:
	void syserr(const char* log)
	{
		fputs(log, stderr);
	}
};

ITextCache* ITextCache::NewTextCache(HDC hSrcDC, uint size, uint codePage)
{
	CTextCache* vpkNewTextCache = new CTextCache;
	if (!vpkNewTextCache->Create(hSrcDC, size, codePage))
	{
		delete vpkNewTextCache;

		return ITextCache::None();
	}

	return vpkNewTextCache;
}

bool Startup()
{
	gs_spkVBRing = IVertexBufferRing::NewVertexBufferRing(Vertex::FVF, sizeof(Vertex)*4, 16);
	return true;
}

void Cleanup()
{
	gs_spkVBRing = NULL;
}

} // text_cache